# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'pyMHF'
copyright = '2025, monkeyman192'
author = 'monkeyman192'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
]

templates_path = ['_templates']
exclude_patterns = []
intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}
add_function_parentheses = False

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ['_static']
html_theme_options = {
    "secondary_sidebar_items": [],
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/monkeyman192/pyMHF",
            "icon": "fa-brands fa-square-github",
            "type": "fontawesome",
        }
    ],
    "back_to_top_button": False,
    "show_nav_level": 0,
    "pygments_light_style": "tango",
    "pygments_dark_style": "monokai"
}



import asyncio
from typing import Union

# This escape sequence is arbitrarily the first 4 digits of Euler's number "e"
# written as bytes from left to right.
ESCAPE_SEQUENCE = b"\x02\x07\x01\x08"
# The "ready" sequence is arbitrarily the first 4 digits of pi written as bytes
# from left to right.
READY_ASK_SEQUENCE = b"\x01\x04\x01\x03"
READY_ACK_SEQUENCE = b"\x06\x02\x09\x05"


class ExecutionEndedException(Exception):
    pass


def custom_exception_handler(loop: asyncio.AbstractEventLoop, context: dict):
    # Simple custom exception handler to stop the loop if an
    # ExecutionEndedException exception is raised.
    exception = context.get("exception")
    if isinstance(exception, ExecutionEndedException):
        loop.stop()


class TerminalProtocol(asyncio.Protocol):
    def __init__(self, message: Union[str, bytes], future):
        super().__init__()
        self.message = message
        self._is_raw = isinstance(self.message, bytes)
        self.future = future
        self._ready_ackd = False

    def connection_made(self, transport):
        self.transport = transport
        if self._is_raw:
            transport.write(self.message)
        else:
            transport.write(self.message.encode())
        if transport.can_write_eof():
            transport.write_eof()

    def data_received(self, data: bytes):
        if data == READY_ACK_SEQUENCE:
            self._ready_ackd = True
            print("READY ACKNOWLEDGED!!!")
        else:
            print(data.decode(), end="")

    def eof_received(self):
        self.transport.close()
        if not self.future.done():
            self.future.set_result(True)

    def connection_lost(self, exc):
        self.transport.close()
        if not self.future.done():
            self.future.set_result(True)
        super().connection_lost(exc)



import ctypes
from concurrent.futures import ThreadPoolExecutor
from enum import Enum


class LoadTypeEnum(Enum):
    INVALID = 0
    SINGLE_FILE = 1
    LIBRARY = 2
    MOD_FOLDER = 3


CWD: str = ""
MODULE_PATH: str = ""
HANDLE = None
MAIN_HWND = None
DPG_HWND = None
PID: int = -1
BINARY_HASH: str = ""
BASE_ADDRESS: int = -1
SIZE_OF_IMAGE: int = -1
CONFIG: dict = {}
EXE_NAME: str = ""
BINARY_PATH: str = ""
LOAD_TYPE: LoadTypeEnum = LoadTypeEnum.INVALID
MOD_SAVE_DIR: str = ""
INCLUDED_ASSEMBLIES: dict[str, str] = {}

_executor: ThreadPoolExecutor = None  # type: ignore


class _GameState:
    def __init__(self):
        self._game_loaded = False

    @property
    def game_loaded(self):
        return self._game_loaded

    @game_loaded.setter
    def game_loaded(self, val: bool):
        # The game can become loaded, but it can't become unloaded...
        if val is True:
            self._game_loaded = val


GameState: _GameState = _GameState()

imports: dict[str, dict[str, ctypes._CFuncPtr]] = {}



import argparse
import os
import os.path as op
import shutil
import subprocess
import sys
from enum import Enum
from importlib.metadata import PackageNotFoundError, entry_points, version
from typing import Optional

import questionary

from .core._types import FUNCDEF  # noqa
from .core.hooking import FuncHook  # noqa
from .core.mod_loader import Mod, ModState  # noqa
from .main import load_mod_file, load_module  # noqa
from .utils.imports import SPHINX_AUTODOC_RUNNING

try:
    __version__ = version("pymhf")
except PackageNotFoundError:
    pass

try:
    from tkinter import Tk, filedialog

    has_tkinter = True
except ModuleNotFoundError:
    has_tkinter = False


class ModeEnum(Enum):
    INVALID = 0
    RUN = 1
    CONFIG = 2
    INIT = 3
    COMMAND = 4
    # DATA = 5


class FolderModeEnum(Enum):
    INVALID = 0
    LIBRARY = 1
    MOD_FOLDER = 2
    NON_LOCAL = 3


def _is_int(val: str) -> bool:
    try:
        int(val)
    except (ValueError, TypeError):
        return False
    return True


def get_folder(title: str, q: questionary.Question, has_tkinter: bool, idir: Optional[str] = None) -> str:
    if has_tkinter:
        return filedialog.askdirectory(initialdir=idir, title=title)
    else:
        return q.ask()


# Pattern shamelessly stolen from the questionary library (which is the actual reason this code has issues in
# the first place!)
if not SPHINX_AUTODOC_RUNNING:
    START_PAUSED = questionary.confirm("Start the game paused?", default=True)
    RUN_GAME = questionary.confirm("Run game?", default=True)
    STEAM_ID_Q = questionary.text("Enter the steam game ID:", validate=_is_int)
    EXE_PATH_Q = questionary.path("Enter the absolute path to the binary:")
    MOD_DIR_Q = questionary.path("Enter the absolute path the mod directory")
    MOD_SAVE_DIR_Q = questionary.path("Enter the absolute path the mod save directory")
    LOG_DIR_Q = questionary.path("Enter the absolute path the logs directory")
    CONTINUE_CONFIGURING_Q = questionary.confirm("Would you like to configure more options?", default=True)

    CFG_OPT_BIN_PATH = "Set binary path"
    CFG_OPT_MOD_PATH = "Set mod directory"
    CFG_OPT_MOD_SAVE_PATH = "Set mod save directory"
    CFG_OPT_STEAM_ID = "Configure steam game id"
    CFG_OPT_LOG_PATH = "Set log directory"
    CFG_OPT_START_PAUSED = "Set game to start paused"
    CONFIG_SELECT_Q = questionary.select(
        "What would you like to configure?",
        choices=[
            CFG_OPT_BIN_PATH,
            CFG_OPT_MOD_PATH,
            CFG_OPT_STEAM_ID,
            CFG_OPT_LOG_PATH,
            CFG_OPT_MOD_SAVE_PATH,
            CFG_OPT_START_PAUSED,
        ],
    )

# This is the name of the config file within the library.
CFG_FILENAME = "pymhf.toml"
# This is the name of the config file which will be sorted in the user data folder.
LOCAL_CFG_FILENAME = "pymhf.local.toml"


def run():
    """Main entrypoint which can be used to run programs with pymhf.
    This will take the first argument as the name of a module which has been installed.
    """
    from .utils.parse_toml import _parse_toml, read_pymhf_settings, write_pymhf_settings

    parser = argparse.ArgumentParser(
        prog="pymhf",
        description="Run the registered plugin",
    )

    parser.add_argument("--version", action="store_true", help="Print the version of pyMHF and exit")

    command_parser = parser.add_subparsers(dest="_command")

    # `run` command parser
    run_parser = command_parser.add_parser(
        "run",
        help=(
            "Run the provided plugin name (if registered), or the provided path to a module or single-file "
            "mod."
        ),
    )
    run_parser.add_argument(
        "plugin_name",
        help=(
            "The name of the installed library to run, or the single-file script to run, or the path to a "
            "folder which contains a library to run locally."
        ),
    )

    config_parser = command_parser.add_parser(
        "config",
        help=(
            "Configure the provided plugin name (if registered), or the provided path to a module or "
            "single-file mod."
        ),
    )
    config_parser.add_argument(
        "-o",
        "--open",
        action="store_true",
        default=False,
        help="Open the directory the local config is stored at.",
    )
    config_parser.add_argument(
        "plugin_name",
        help=(
            "The name of the installed library to run, or the single-file script to run, or the path to a "
            "folder which contains a library to run locally."
        ),
    )

    init_parser = command_parser.add_parser(
        "init", help="[NOT YET IMPLEMENTED] Initialise a new empty project under the given name."
    )
    init_parser.add_argument(
        "plugin_name",
        help=(
            "The name of the installed library to run, or the single-file script to run, or the path to a "
            "folder which contains a library to run locally."
        ),
    )

    cmd_parser = command_parser.add_parser("cmd", help="Perform a registered command for the library")
    cmd_parser.add_argument(
        "command",
        help=(
            "The command to be run by the specified library. These are registered in the "
            "[project.entry-points.pymhfcmd] section of the pyproject.toml file."
        ),
    )
    cmd_parser.add_argument(
        "plugin_name",
        help=(
            "The name of the installed library to run, or the path to a folder which contains a library to "
            "run locally."
        ),
    )

    # data_parser = command_parser.add_parser(
    #     "data", help="Configure the data for the provided plugin name (if registered), or the provided path"
    #     " to a module."
    # )
    # data_parser.add_argument(
    #     "plugin_name",
    #     help=(
    #         "The name of the installed library to run, or the path to a folder which contains a library to "
    #         "run locally."
    #     ),
    # )
    # data_parser.add_argument(
    #     "-H",
    #     "--hash",
    #     action="store_true",
    #     help="Get the current binary hash",
    # )

    args, extras = parser.parse_known_args()  # noqa

    if args.version:
        print(__version__)
        sys.exit(0)

    # TODO: The extras can be passed to the registered library in the future.

    try:
        plugin_name: str = args.plugin_name
    except AttributeError:
        parser.error("No command provided. Please provide a valid command. See `pymhf --help` for details.")
    command = args._command
    if command == "run":
        mode = ModeEnum.RUN
    elif command == "config":
        mode = ModeEnum.CONFIG
    elif command == "init":
        mode = ModeEnum.INIT
    elif command == "cmd":
        mode = ModeEnum.COMMAND
    # elif command == "data":
    #     mode = ModeEnum.DATA
    else:
        mode = ModeEnum.INVALID
    standalone = False
    local = False

    if op.isfile(plugin_name) and op.exists(plugin_name):
        # In this case we are running in stand-alone mode
        standalone = True

    if op.isdir(plugin_name) and op.exists(plugin_name):
        # In this case we are running a library directly from pymhf.
        # This can be done for two reasons... We either want to actually run it, or we are configuring it.
        local = True

    if standalone:
        if mode != ModeEnum.RUN:
            print("Only the `run` command can be used with single-file mods.")
            if mode == ModeEnum.CONFIG:
                print(
                    "To configure a single-file mod, change the values directly in the inline script "
                    "metadata. See https://monkeyman192.github.io/pyMHF/docs/settings.html"
                )
            return
        load_mod_file(plugin_name)
        return

    cmd_command = None
    if mode == ModeEnum.COMMAND:
        cmd_command = args.command

    folder_mode = FolderModeEnum.INVALID

    if local:
        # Check to see if the folder we are pointing to has a pyproject.toml file, or a pymhf.toml file.
        # If it's a pyproject.toml, then we are running a library directly.
        # If it's a pymhf.toml, then it's a mod folder.
        # Parse the pyproject.toml file to get some info...
        # We'll give precedency to the pyproject.toml file.
        local_plugin_dir = op.realpath(plugin_name)
        pyproject_toml = op.join(local_plugin_dir, "pyproject.toml")
        pymhf_toml = op.join(local_plugin_dir, CFG_FILENAME)
        if op.exists(pyproject_toml):
            folder_mode = FolderModeEnum.LIBRARY
        elif op.exists(pymhf_toml):
            folder_mode = FolderModeEnum.MOD_FOLDER

        if folder_mode == FolderModeEnum.INVALID:
            print(
                f"Error: No pyproject.toml or pymhf.toml file in the directory {local_plugin_dir}."
                "Please ensure the target project has one of these."
            )
            return
        if folder_mode == FolderModeEnum.LIBRARY:
            settings = _parse_toml(pyproject_toml, False)
            plugin_name = None
            if (
                project_cfg := settings.get("project", {}).get("entry-points", {}).get("pymhflib")
            ) is not None:
                plugin_name = list(project_cfg.keys())[0]
                if mode == ModeEnum.COMMAND:
                    if (
                        command_cfg := settings.get("project", {}).get("entry-points", {}).get("pymhfcmd")
                    ) is not None:
                        from .core.importing import import_file

                        if (resolved_command := command_cfg.get(cmd_command)) is not None:
                            mod = import_file(op.join(local_plugin_dir, plugin_name))
                            cmd = getattr(mod, resolved_command, None)
                            if cmd is not None:
                                cmd(extras)
                                return
                            else:
                                print(
                                    f"Command {resolved_command!r} cannot be found in module {plugin_name}. "
                                    "Please ensure the function exists and can be run from this module."
                                )
                                return
                        else:
                            print(f"No registered command {cmd_command!r}. Please check your configuration.")
                            return
                    else:
                        print(
                            "No [project.entry-points.pymhfcmd] section found in the pyproject.toml. "
                            "Please create one to register and run custom commands."
                        )
                        return
        elif folder_mode == FolderModeEnum.MOD_FOLDER:
            # Parse the pymhf.toml file to get some details about what we are running.
            settings = _parse_toml(pymhf_toml, False)
            # We'll set the plugin name as the name of the folder
            plugin_name = op.basename(local_plugin_dir)
        if plugin_name is None:
            print(
                f"Cannot determine the project at the path specified {plugin_name}.\n"
                "Please ensure the pyproject.toml file has the [project.entry-points.pymhflib] entry."
                "This process will not continue."
            )
            return
    else:
        folder_mode = FolderModeEnum.NON_LOCAL

    # Get the location of app data, then construct the expected folder name.
    appdata_data = os.environ.get("APPDATA", op.expanduser("~"))
    if appdata_data == "~":
        # In this case the APPDATA environment variable isn't set and ~ also fails to resolve.
        # Raise a error and stop.
        print("Critical Error: Cannot find user directory. Ensure APPDATA environment variable is set")
        exit()
    cfg_folder = op.join(appdata_data, "pymhf", plugin_name)
    os.makedirs(cfg_folder, exist_ok=True)

    # PROCESS:
    """
    1. Check that the config file is in the correct location. If not, prompt the user to configure it.
    2. if no `--config` argument passed, then we just run the program as intented.
    3. If `--config` is provided, then we don't run the program and instead configure it.
    4. The arguments for config are the keys (this will need to be documented extensively) followed by values
        and separated by an `=`.
    """

    initial_config = False

    if not local:
        eps = entry_points()
        # This check is to ensure compatibility with multiple versions of python as the code 3.10+ isn't
        # backward compatible.
        if isinstance(eps, dict):
            pymhf_entry_points = eps.get("pymhflib", [])
        else:
            pymhf_entry_points = eps.select(group="pymhflib")
        if mode == ModeEnum.COMMAND:
            if isinstance(eps, dict):
                pymhf_commands = eps.get("pymhfcmd", [])
            else:
                pymhf_commands = eps.select(group="pymhfcmd")

            if not pymhf_commands:
                print(
                    "No [project.entry-points.pymhfcmd] section found in the pyproject.toml. "
                    "Please create one to register and run custom commands."
                )

        required_lib = None
        resolved_command = None

        for ep in pymhf_entry_points:
            if ep.name.lower() == plugin_name.lower():
                required_lib = ep
        if mode == ModeEnum.COMMAND:
            for cmd in pymhf_commands:
                # See if the library has a check command defined.
                if cmd.name.lower() == cmd_command:
                    resolved_command = cmd.value

        if required_lib is None:
            print(
                f"Cannot find {plugin_name} as an installed plugin. "
                "Please ensure it has been installed and try again"
            )
            return

        module_dir = op.dirname(required_lib.load().__file__)

        if mode == ModeEnum.COMMAND:
            if resolved_command is not None:
                # We can use importlib here since the library has to be installed if we have made it here.
                import importlib

                mod = importlib.import_module(plugin_name)
                cmd = getattr(mod, resolved_command, None)
                if cmd is not None:
                    cmd(extras)
                    return
                else:
                    print(
                        f"Command {resolved_command!r} cannot be found in module {plugin_name}. "
                        "Please ensure the function exists and can be run from this module."
                    )
                    return
            else:
                print(f"No registered command {cmd_command!r}. Please check your configuration.")
                return

    else:
        if folder_mode == FolderModeEnum.LIBRARY:
            module_dir = op.join(local_plugin_dir, plugin_name)
        else:
            module_dir = local_plugin_dir

    cfg_file = op.join(module_dir, CFG_FILENAME)
    config_progress_file = op.join(cfg_folder, ".config_in_progress")
    if not op.exists(cfg_file):
        print(
            f"Cannot find `{CFG_FILENAME}` for {plugin_name}! This is likely an error on the maintainers' "
            "behalf.\nCannot continue loading until this is fixed."
        )
        return
    else:
        dst = op.join(cfg_folder, LOCAL_CFG_FILENAME)
        if not op.exists(dst) or op.exists(config_progress_file):
            # In this case we can prompt the user to enter the config values which need to be changed.
            initial_config = True

    if initial_config:
        local_config = {"local_config": {}}
        if not has_tkinter:
            print(
                "tkinter cannot be found. Please ensure it's installed as part of your python install.\n"
                "Falling back to the command-line method of providing an input."
            )
        # Copy the config file to the appdata directory.
        shutil.copyfile(cfg_file, dst)
        # Write the file which indicates we are in progress.
        with open(config_progress_file, "w") as f:
            f.write("")

        if folder_mode != FolderModeEnum.MOD_FOLDER:
            # Modify some of the values in the config file, allowing the user to enter the values they want.
            # We only do this if we are not dealing with a mod folder...
            if has_tkinter:
                root = Tk()
                root.withdraw()
            if (
                mod_folder := get_folder("Select folder where mods are located", MOD_DIR_Q, has_tkinter)
            ) is not None:
                local_config["local_config"]["mod_dir"] = mod_folder
            else:
                return None

        # Write the config back and then delete the temporary file only once everything is ok.
        write_pymhf_settings(local_config, dst)
        os.remove(config_progress_file)
        initial_config = False
    elif mode == ModeEnum.CONFIG:
        if args.open:
            print(f"Opening {cfg_folder!r} and exiting")
            subprocess.call(f'explorer "{cfg_folder}"')
            return
        local_config = read_pymhf_settings(dst)
        pymhf_settings = local_config["local_config"]
        keep_going = True
        if has_tkinter:
            root = Tk()
            root.withdraw()
        while keep_going:
            config_choice = CONFIG_SELECT_Q.ask()
            if config_choice == CFG_OPT_BIN_PATH:
                if (exe_path := EXE_PATH_Q.ask()) is not None:
                    pymhf_settings["exe_path"] = exe_path
                    if "steam_gameid" in pymhf_settings:
                        del pymhf_settings["steam_gameid"]
                else:
                    return
            elif config_choice == CFG_OPT_STEAM_ID:
                if (steam_id := STEAM_ID_Q.ask()) is not None:
                    pymhf_settings["steam_gameid"] = steam_id
                    if "exe_path" in pymhf_settings:
                        del pymhf_settings["exe_path"]
                else:
                    return
            elif config_choice == CFG_OPT_MOD_PATH:
                if (
                    mod_dir := get_folder("Select folder where mods can be found", MOD_DIR_Q, has_tkinter)
                ) is not None:
                    pymhf_settings["mod_dir"] = mod_dir
                else:
                    return None
            elif config_choice == CFG_OPT_START_PAUSED:
                if (start_paused := START_PAUSED.ask()) is not None:
                    pymhf_settings["start_paused"] = start_paused
                else:
                    return
            elif config_choice == CFG_OPT_LOG_PATH:
                if (
                    log_dir := get_folder("Select folder where logs are placed", LOG_DIR_Q, has_tkinter)
                ) is not None:
                    pymhf_settings["log_dir"] = log_dir
                else:
                    return None
            elif config_choice == CFG_OPT_MOD_SAVE_PATH:
                if (
                    msd := get_folder("Select folder where mod saves are placed", MOD_SAVE_DIR_Q, has_tkinter)
                ) is not None:
                    pymhf_settings["mod_save_dir"] = msd
                else:
                    return None
            elif config_choice is None:
                keep_going = False

            keep_going = CONTINUE_CONFIGURING_Q.ask()
            if keep_going is None:
                return
        local_config["local_config"] = pymhf_settings
        write_pymhf_settings(local_config, dst)

        if not RUN_GAME.ask():
            return
    # Final step: If we have either finished the initial config, or we didn't want to config at all, run the
    # binary.
    if not initial_config:
        load_module(plugin_name, module_dir)



import argparse
import logging
import logging.handlers
import os
import os.path as op
import pickle
import select
import socketserver
import struct
import time

import pymhf

# Logo generated using https://patorjk.com/software/taag/
# Options:
# Font: Varsity
# Character Width: Full
# Character Height: Full
LOGO = """
 _______    ____  ____   ____    ____   ____  ____   ________  
|_   __ \  |_  _||_  _| |_   \  /   _| |_   ||   _| |_   __  | 
  | |__) |   \ \  / /     |   \/   |     | |__| |     | |_ \_| 
  |  ___/     \ \/ /      | |\  /| |     |  __  |     |  _|    
 _| |_        _|  |_     _| |_\/_| |_   _| |  | |_   _| |_     
|_____|      |______|   |_____||_____| |____||____| |_____|    
                                                               
"""


# NB: This code is mostly taken from the python stdlib docs.


class LogRecordStreamHandler(socketserver.StreamRequestHandler):
    """Handler for a streaming logging request.

    This basically logs the record using whatever logging policy is
    configured locally.
    """

    def handle(self):
        """
        Handle multiple requests - each expected to be a 4-byte length,
        followed by the LogRecord in pickle format. Logs the record
        according to whatever policy is configured locally.
        """
        try:
            while True:
                chunk = self.connection.recv(4)
                if len(chunk) < 4:
                    break
                slen = struct.unpack(">L", chunk)[0]
                chunk = self.connection.recv(slen)
                while len(chunk) < slen:
                    chunk = chunk + self.connection.recv(slen - len(chunk))
                obj = self.unPickle(chunk)
                record = logging.makeLogRecord(obj)
                self.handleLogRecord(record)
        except ConnectionResetError:
            return

    def unPickle(self, data):
        return pickle.loads(data)

    def handleLogRecord(self, record):
        # If a name is specified, we use the named logger rather than the one
        # implied by the record.
        if self.server.logname is not None:
            name = self.server.logname
        else:
            name = record.name
        logger = logging.getLogger(name)
        # N.B. EVERY record gets logged. This is because Logger.handle
        # is normally called AFTER logger-level filtering. If you want
        # to do filtering, do it at the client end to save wasting
        # cycles and network bandwidth!
        logger.handle(record)


class LogRecordSocketReceiver(socketserver.ThreadingTCPServer):
    """
    Simple TCP socket-based logging receiver suitable for testing.
    """

    allow_reuse_address = True

    def __init__(
        self,
        host="localhost",
        port=logging.handlers.DEFAULT_TCP_LOGGING_PORT,
        handler=LogRecordStreamHandler,
    ):
        socketserver.ThreadingTCPServer.__init__(self, (host, port), handler)
        self.abort = 0
        self.timeout = 1
        self.logname = None

    def serve_until_stopped(self):
        abort = 0
        try:
            while not abort:
                rd, _, _ = select.select([self.socket.fileno()], [], [], self.timeout)
                if rd:
                    self.handle_request()
                abort = self.abort
        except KeyboardInterrupt:
            print("Ending logging server...")


def main(path: str):
    formatter = logging.Formatter("%(asctime)s %(name)-15s %(levelname)-6s %(message)s")
    logdir = op.join(path, "..", "logs")
    os.makedirs(logdir, exist_ok=True)
    # TODO: Need to make this strip the ANSI escape chars from the written log
    file_handler = logging.FileHandler(
        op.join(logdir, f"pymhf-{time.strftime('%Y%m%dT%H%M%S')}.log"), encoding="utf-8"
    )
    file_handler.setFormatter(formatter)
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    root_logger = logging.getLogger()
    root_logger.addHandler(file_handler)
    root_logger.addHandler(stream_handler)
    tcpserver = LogRecordSocketReceiver()
    print(LOGO)
    print(f"Version: {pymhf.__version__}")
    print("Logger waiting for backend process... Please wait...")
    tcpserver.serve_until_stopped()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("path", help="Path to place the log files")
    args = parser.parse_args()
    main(args.path)



from typing import Any, Callable, Optional

from pymhf.gui.protocols import ButtonProtocol, ComboBoxProtocol, VariableProtocol, VariableType


def gui_button(text: str):
    def inner(func) -> ButtonProtocol:
        func._is_button = True
        func._button_text = text
        return func

    return inner


def gui_combobox(text: str, items: Optional[list[str]] = None):
    if items is None:
        items = []

    def inner(func) -> ComboBoxProtocol:
        func._is_combobox = True
        func._combobox_text = text
        func._items = items
        return func

    return inner


def no_gui(cls):
    """Mark the mod as not requiring a tab in the gui."""
    setattr(cls, "_no_gui", True)
    return cls


class gui_variable:
    @staticmethod
    def _set_default_attributes(func: Callable[..., Any], label: Optional[str] = None):
        func._is_variable = True
        func._label_text = label
        func._has_setter = False

    @staticmethod
    def _clean_extra_args(args: dict):
        """Remove any keywords from the args so that they don't override the values we need to provide."""
        args.pop("tag", None)
        args.pop("source", None)
        args.pop("user_data", None)
        args.pop("callback", None)
        args.pop("use_internal_label", None)

    @classmethod
    def INTEGER(cls, label: Optional[str] = None, **extra_args):
        """Create an integer entry field which can take extra arguments.
        To see what extra arguments are available, see the DearPyGUI documentation:
        https://dearpygui.readthedocs.io/en/latest/reference/dearpygui.html#dearpygui.dearpygui.add_input_int
        """

        def inner(func: Callable[..., Any]) -> VariableProtocol:
            gui_variable._set_default_attributes(func, label)
            func._variable_type = VariableType.INTEGER
            gui_variable._clean_extra_args(extra_args)
            func._extra_args = extra_args
            return func

        return inner

    @classmethod
    def STRING(cls, label: Optional[str] = None, **extra_args):
        """Create an string entry field which can take extra arguments.
        To see what extra arguments are available, see the DearPyGUI documentation:
        https://dearpygui.readthedocs.io/en/latest/reference/dearpygui.html#dearpygui.dearpygui.add_input_text
        """

        def inner(func: Callable[..., Any]) -> VariableProtocol:
            gui_variable._set_default_attributes(func, label)
            func._variable_type = VariableType.STRING
            gui_variable._clean_extra_args(extra_args)
            func._extra_args = extra_args
            return func

        return inner

    @classmethod
    def FLOAT(cls, label: Optional[str] = None, **extra_args):
        """Create an float entry field which can take extra arguments.
        To see what extra arguments are available, see the DearPyGUI documentation:
        https://dearpygui.readthedocs.io/en/latest/reference/dearpygui.html#dearpygui.dearpygui.add_input_double
        """

        def inner(func: Callable[..., Any]) -> VariableProtocol:
            gui_variable._set_default_attributes(func, label)
            func._variable_type = VariableType.FLOAT
            gui_variable._clean_extra_args(extra_args)
            func._extra_args = extra_args
            return func

        return inner

    @classmethod
    def BOOLEAN(cls, label: Optional[str] = None, **extra_args):
        """Create an boolean entry field in the form of a checkbox which can take extra arguments.
        To see what extra arguments are available, see the DearPyGUI documentation:
        https://dearpygui.readthedocs.io/en/latest/reference/dearpygui.html#dearpygui.dearpygui.add_checkbox
        """

        def inner(func: Callable[..., Any]) -> VariableProtocol:
            gui_variable._set_default_attributes(func, label)
            func._variable_type = VariableType.BOOLEAN
            gui_variable._clean_extra_args(extra_args)
            func._extra_args = extra_args
            return func

        return inner


INTEGER = gui_variable.INTEGER
BOOLEAN = gui_variable.BOOLEAN
STRING = gui_variable.STRING
FLOAT = gui_variable.FLOAT



import ctypes
import struct
from datetime import datetime
from logging import getLogger
from typing import Iterable, Optional, TypedDict, Union

import dearpygui.dearpygui as dpg
from pymem.ressources.structure import MEMORY_STATE

from pymhf.utils.winapi import MemoryInfo, QueryAddress

logger = getLogger("hexView")

BITS = struct.calcsize("P") * 8

COLUMNS = 0x10
MAX_DATA_CACHE_SIZE = 0x100  # 256
MAX_FRAME_SIZE = 0x1000  # 4kb of memory
# Combination of the above gives us up to 1Mb of memory cached.

INITIAL_DATA = b"\x00" * MAX_FRAME_SIZE


class CachedMemoryData(TypedDict):
    base_address: int
    address: int
    data: bytes
    timestamp: datetime


def chunks(lst: Iterable, n: int):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


class HexView:
    def __init__(self, parent: Union[str, int]):
        self.parent = parent
        self.prev_selected_coords = set()
        self.curr_selected_coords = set()
        self.data: bytes = INITIAL_DATA
        self._select_bytes_start_addr: int = 0
        self._selected_bytes: bytes = b""
        self._open_file = None
        self.base_address: int = 0

        # Basic forward and backward address stack
        self._address_stack = []
        self._address_stack_location: int = 0

        # Stored memory data
        self._memory_cache: dict[str, CachedMemoryData] = {}
        self._memory_cache_repeated_tags: dict[str, int] = {}
        self._selected_snapshot_tag = None

    def add_snapshot(self, address: int, size: Optional[int] = None, tag: Optional[str] = None) -> bool:
        """Add a snapshot of the data at the specified address to the hex viewer. These memory snapshots will
        be viewable later from a drop down in the hex viewer.

        Parameters
        ----------
        address:
            The absolute memory address to take a snapshot at.
            *Note*: The start address of the region will be aligned to a 0x10 byte boundary for convenience.
        size:
            The size of the region to take a snapshot of. Note that if this is larger than 0x1000 (4kb), it
            will be rounded down to 0x1000.
            *Note*: It is recommended to set this value just a bit larger than the expected size of the data.
        tag:
            An optional tag which can be used to identify the snapshot in the selector in the GUI.
            If not provided it will default to the hex representation of the provided memory address.
            *Note*: If the tag already exists, it won't be overwritten but will be suffixed by a number which
            automatically increments (this includes auto-generated tags as per above.)

        Returns
        -------
        bool:
            If a memory snapshot was sucessfully taken at the provided address and size this will return True.
            If the memory at the provided address was unable to be read or captured this will return False.
        """
        addr_data = self._inspect_memory_region(address)
        if addr_data is None:
            return False

        base_address, data = self._load_data_from_address(address, addr_data, size)

        # Either generate a tag, or use the provided one.
        # If the provided one already exists or the generated one already exists, then we'll add a number
        # which gets incremented to the end of it.
        _tag = tag
        if tag is not None:
            if tag in self._memory_cache:
                if tag not in self._memory_cache_repeated_tags:
                    self._memory_cache_repeated_tags[tag] = 1
                else:
                    self._memory_cache_repeated_tags[tag] += 1
                _tag = f"{tag}-{self._memory_cache_repeated_tags[tag]}"
        else:
            _tag = hex(address)
            if _tag in self._memory_cache:
                if _tag not in self._memory_cache_repeated_tags:
                    self._memory_cache_repeated_tags[_tag] = 1
                else:
                    self._memory_cache_repeated_tags[_tag] += 1
                _tag = f"{_tag}-{self._memory_cache_repeated_tags[_tag]}"

        self._memory_cache[_tag] = {
            "base_address": base_address,
            "address": address,
            "data": data,
            "timestamp": datetime.now(),
        }

        # Update the list of items in the combobox
        dpg.configure_item(
            "_snapshot_dropdown",
            items=list(self._memory_cache.keys()),
        )

        logger.debug(
            f"Added snapshot for the memory region 0x{base_address:X} -> 0x{base_address + len(data):X} "
            f"with the tag {_tag!r}"
        )

        return True

    def load_address(self, address: int, size: Optional[int] = None) -> bool:
        """Load the provided address into the hex viewer.

        Parameters
        ----------
        address:
            The absolute memory address to take a snapshot at.
            *Note*: The start address of the region will be aligned to a 0x10 byte boundary for convenience.
        size:
            The size of the region to take a snapshot of. Note that if this is larger than 0x1000 (4kb), it
            will be rounded down to 0x1000.
            *Note*: It is recommended to set this value just a bit larger than the expected size of the data.

        Returns
        -------
        bool:
            If a memory snapshot was sucessfully taken at the provided address and size this will return True.
            If the memory at the provided address was unable to be read or captured this will return False.
        """
        addr_data = self._inspect_memory_region(address)
        if addr_data is None:
            return False

        # Cache the previous data size so that we may optimize the redraw of the table if they are the same
        # size.
        prev_size = len(self.data)

        self.base_address, self.data = self._load_data_from_address(address, addr_data, size)
        new_size = len(self.data)

        logger.debug(f"Refreshing table to show {len(self.data)} bytes from 0x{self.base_address:X}")
        # Clear and re-populate the table.
        if new_size == prev_size:
            self._refresh_table()
        else:
            self._clear_table()
            self._populate_table()
        # Select the currently selected value.
        self._select_coord(0, address % 0x10)

        # Update the address stack
        if self._address_stack_location == len(self._address_stack) - 1:
            # If we're at the end, just append the value.
            self._address_stack.append(self.base_address)
        else:
            # If we're not at the end, then we need to remove the next addresses in the stack and push this
            # to the end.
            self._address_stack = self._address_stack[: self._address_stack_location + 1] + [
                self.base_address
            ]
        self._address_stack_location = len(self._address_stack) - 1

        return True

    def _change_address(self, sender, app_data):
        val = int(app_data, 0x10)
        self.load_address(val)

    def _change_selection_size(self, sender, app_data, user_data):
        if app_data == 2:
            dpg.configure_item("size_2_values", show=True)
            dpg.configure_item("size_4_values", show=False)
            dpg.configure_item("size_8_values", show=False)
            dpg.configure_item("size_inv_values", show=False)
        elif app_data == 4:
            dpg.configure_item("size_2_values", show=False)
            dpg.configure_item("size_4_values", show=True)
            dpg.configure_item("size_8_values", show=False)
            dpg.configure_item("size_inv_values", show=False)
        elif app_data == 8:
            dpg.configure_item("size_2_values", show=False)
            dpg.configure_item("size_4_values", show=False)
            dpg.configure_item("size_8_values", show=True)
            dpg.configure_item("size_inv_values", show=False)
        else:
            dpg.configure_item("size_2_values", show=False)
            dpg.configure_item("size_4_values", show=False)
            dpg.configure_item("size_8_values", show=False)
            dpg.configure_item("size_inv_values", show=True)

    def _clear_table(self):
        dpg.delete_item("main_table", children_only=True, slot=1)

    def _clicked_on(self, sender, app_data):
        if dpg.get_value(app_data[1]):
            dpg.configure_item("option_popup", show=True)

    def _delete_snapshot(self):
        curr_idx = list(self._memory_cache.keys()).index(self._selected_snapshot_tag)
        _data = self._memory_cache.pop(self._selected_snapshot_tag)
        dpg.configure_item(
            "_snapshot_dropdown",
            items=list(self._memory_cache.keys()),
        )
        if curr_idx > 0:
            new_snapshot_tag = list(self._memory_cache.keys())[curr_idx - 1]
            self._load_snapshot(new_snapshot_tag)
            dpg.set_value("_snapshot_dropdown", new_snapshot_tag)
        elif len(self._memory_cache) > 0:
            new_snapshot_tag = list(self._memory_cache.keys())[0]
            self._load_snapshot(new_snapshot_tag)
            dpg.set_value("_snapshot_dropdown", new_snapshot_tag)
        else:
            # no snapshots left...
            self.base_address, self.data = 0, INITIAL_DATA
            if len(_data) == MAX_FRAME_SIZE:
                self._refresh_table()
            else:
                self._clear_table()
                self._populate_table()
            self._select_coord(0, 0)

            dpg.configure_item("_snapshot_details_group", show=False)
            dpg.configure_item("_snapshot_details_text", show=False)
            dpg.set_value("_snapshot_dropdown", None)

    def _inspect_memory_region(self, address: int) -> Optional[MemoryInfo]:
        """Inspect a memory address and determine if its ok to store or display."""
        try:
            addr_data = QueryAddress(address)
        except ValueError as e:
            logger.error(f"Unable to load address 0x{address:X}: {e.args}")
            return None
        logger.debug(
            f"Info about 0x{address:X}:\n"
            f"\tBaseAddress: 0x{addr_data.BaseAddress:X}\n"
            f"\tAllocationBase: 0x{addr_data.AllocationBase:X}\n"
            f"\tAllocationProtect: 0x{addr_data.AllocationProtect:X}\n"
            f"\tRegionSize: 0x{addr_data.RegionSize:X}\n"
            f"\tState: 0x{addr_data.State:X}\n"
            f"\tProtect: 0x{addr_data.Protect:X}\n"
            f"\tType: 0x{addr_data.Type:X}\n"
        )

        if not addr_data.State == MEMORY_STATE.MEM_COMMIT:
            logger.error(
                f"Cannot load memory region containing address 0x{address:X}. "
                f"Memory state is {addr_data.State}"
            )
            return None

        return addr_data

    def _load_data_from_address(
        self,
        address: int,
        addr_data: MemoryInfo,
        size: Optional[int] = None,
    ) -> tuple[int, bytes]:
        """Load the data from the address."""
        # Align the requested address to the nearest 0x10 byte boundary before-hand, and then determine how
        # many bytes are afterwards.
        address_0x10_aligned = address - (address % 0x10)
        after_address_size = addr_data.RegionSize - (address_0x10_aligned - addr_data.BaseAddress)

        # We'll only read at most MAX_FRAME_SIZE bytes to keep it sensible, but we may not be able to read
        # more anyway, or may only request less.
        if size is None:
            _size = min(MAX_FRAME_SIZE, after_address_size)
        else:
            _size = min(MAX_FRAME_SIZE, min(size, after_address_size))

        # Finally, read the data at the address.
        _data = (ctypes.c_char * _size).from_address(address_0x10_aligned)
        return address_0x10_aligned, bytes(bytearray(_data))

    def _load_snapshot(self, tag):
        """Load snapshot data from the cache."""
        cache_data = self._memory_cache.get(tag)
        if cache_data is None:
            return

        prev_size = len(self.data)

        self.data = cache_data["data"]
        self.base_address = cache_data["base_address"]

        new_size = len(self.data)

        # Clear and re-populate the table.
        if new_size == prev_size:
            self._refresh_table()
        else:
            self._clear_table()
            self._populate_table()
        self._select_coord(0, cache_data.get("address", 0) % 0x10)

        dpg.configure_item("_snapshot_details_group", show=True)
        dpg.configure_item("_snapshot_details_text", show=True)
        dpg.configure_item(
            "_snapshot_details_timestamp",
            default_value=f"    Timestamp: {cache_data['timestamp']}",
        )

        self._selected_snapshot_tag = tag

    def _populate_table(self):
        data_chunks = list(chunks(self.data, COLUMNS))

        with dpg.table_row(parent="main_table"):
            with dpg.table(
                header_row=False,
                policy=dpg.mvTable_SizingFixedSame,
                no_host_extendX=True,
                borders_innerH=True,
                no_pad_innerX=True,
                no_pad_outerX=True,
                clipper=True,
            ):
                dpg.add_table_column()
                for i in range(len(data_chunks)):
                    with dpg.table_row():
                        dpg.add_selectable(
                            label=f"0x{self.base_address + COLUMNS * i:X}",
                            enabled=False,
                            tag=f"addr_{i}",
                        )

            with dpg.table(
                header_row=False,
                policy=dpg.mvTable_SizingFixedSame,
                no_host_extendX=True,
                borders_innerH=True,
                borders_outerH=True,
                borders_innerV=True,
                borders_outerV=True,
                clipper=True,
            ):
                for _ in range(COLUMNS):
                    dpg.add_table_column(no_resize=True)
                for i, chunk in enumerate(data_chunks):
                    with dpg.table_row():
                        for j, char_ in enumerate(chunk):
                            dpg.add_selectable(
                                label=f"{char_:02X}",
                                callback=self._select_byte,
                                tag=f"byte_{i}-{j}",
                                user_data=(i, j),
                            )
                            dpg.bind_item_handler_registry(f"byte_{i}-{j}", "byte_handler")

            with dpg.table(
                header_row=False,
                policy=dpg.mvTable_SizingFixedSame,
                no_host_extendX=True,
                borders_innerH=True,
                borders_outerH=True,
                borders_innerV=True,
                borders_outerV=True,
                clipper=True,
            ):
                for _ in range(COLUMNS):
                    dpg.add_table_column(no_resize=True)
                for i, chunk in enumerate(data_chunks):
                    with dpg.table_row():
                        for j, char_ in enumerate(chunk):
                            char_conv = chr(char_)
                            if not str.isprintable(char_conv):
                                char_conv = " "
                            dpg.add_selectable(
                                label=char_conv,
                                callback=self._select_str,
                                tag=f"str_{i}-{j}",
                                user_data=(i, j),
                            )

    def _popup_follow_pointer(self):
        # Get the bytes which correspond to the ptr value and then cast to an int.
        start_addr = self._select_bytes_start_addr - self.base_address
        ptr_bytes = self.data[start_addr : start_addr + BITS // 8]
        ptr = int.from_bytes(ptr_bytes, byteorder="little", signed=False)
        dpg.configure_item("option_popup", show=False)
        self.load_address(ptr)

    def _refresh_table(self):
        """Refresh the table with new data instead of removing the old data."""
        data_chunks = list(chunks(self.data, COLUMNS))
        for i, chunk in enumerate(data_chunks):
            dpg.configure_item(f"addr_{i}", label=f"0x{self.base_address + COLUMNS * i:X}")
            for j, char_ in enumerate(chunk):
                char_conv = chr(char_)
                if not str.isprintable(char_conv):
                    char_conv = " "
                dpg.configure_item(f"byte_{i}-{j}", label=f"{char_:02X}")
                dpg.configure_item(f"str_{i}-{j}", label=char_conv)

    def _select_byte(self, sender, app_data, user_data):
        self._select_coord(*user_data)

    def _select_coord(self, i: int, j: int):
        idx = i * COLUMNS + j
        selection_size = dpg.get_value("selection_size")
        self.prev_selected_coords = set(self.curr_selected_coords)
        self.curr_selected_coords = set()
        self._selected_bytes = self.data[idx : idx + selection_size]
        self._select_bytes_start_addr = self.base_address + idx
        dpg.set_value("address_input", hex(self._select_bytes_start_addr)[2:])
        if selection_size == 2:
            dpg.set_value(
                "_select_bytes_int16",
                str(int.from_bytes(self._selected_bytes, byteorder="little", signed=False)),
            )
            dpg.set_value(
                "_select_bytes_uint16",
                str(int.from_bytes(self._selected_bytes, byteorder="little", signed=True)),
            )
        elif selection_size == 4:
            dpg.set_value(
                "_select_bytes_int32",
                str(int.from_bytes(self._selected_bytes, byteorder="little", signed=False)),
            )
            dpg.set_value(
                "_select_bytes_uint32",
                str(int.from_bytes(self._selected_bytes, byteorder="little", signed=True)),
            )
        elif selection_size == 8:
            dpg.set_value(
                "_select_bytes_int64",
                str(int.from_bytes(self._selected_bytes, byteorder="little", signed=False)),
            )
            dpg.set_value(
                "_select_bytes_uint64",
                str(int.from_bytes(self._selected_bytes, byteorder="little", signed=True)),
            )
        for s in range(selection_size):
            x, y = ((idx + s) // COLUMNS, (idx + s) % COLUMNS)
            self.curr_selected_coords.add((x, y))
            try:
                dpg.set_value(f"byte_{x}-{y}", True)
                dpg.set_value(f"str_{x}-{y}", True)
            except SystemError:
                pass
        for coord in self.prev_selected_coords - self.curr_selected_coords:
            x, y = coord
            try:
                dpg.set_value(f"byte_{x}-{y}", False)
                dpg.set_value(f"str_{x}-{y}", False)
            except SystemError:
                pass

    def _select_snapshot(self, sender, app_data):
        self._load_snapshot(app_data)

    def _select_str(self, sender, app_data, user_data):
        self._select_coord(*user_data)

    def _setup(self):
        with dpg.theme() as table_theme:
            with dpg.theme_component(dpg.mvTable):
                dpg.add_theme_style(
                    dpg.mvStyleVar_SelectableTextAlign,
                    0.5,
                    0.5,
                    category=dpg.mvThemeCat_Core,
                )
                dpg.add_theme_color(
                    dpg.mvThemeCol_HeaderHovered,
                    (255, 0, 0, 100),
                    category=dpg.mvThemeCat_Core,
                )
                dpg.add_theme_color(
                    dpg.mvThemeCol_HeaderActive,
                    (0, 255, 0, 100),
                    category=dpg.mvThemeCat_Core,
                )
                dpg.add_theme_color(dpg.mvThemeCol_Header, (0, 0, 255, 100), category=dpg.mvThemeCat_Core)

        with dpg.value_registry():
            # Options
            dpg.add_string_value(tag="address_input", default_value="10")
            dpg.add_int_value(tag="selection_size", default_value=4)
            # Selected value values
            dpg.add_string_value(tag="_select_bytes_int16", default_value="N/A")
            dpg.add_string_value(tag="_select_bytes_uint16", default_value="N/A")
            dpg.add_string_value(tag="_select_bytes_int32", default_value="N/A")
            dpg.add_string_value(tag="_select_bytes_uint32", default_value="N/A")
            dpg.add_string_value(tag="_select_bytes_int64", default_value="N/A")
            dpg.add_string_value(tag="_select_bytes_uint64", default_value="N/A")

        self._setup_rightclick_handlers()

        # with dpg.handler_registry(tag="table_keypress_handlers"):
        #     dpg.add_key_press_handler(dpg.mvKey_Down, callback=self._press_down)
        #     dpg.add_key_press_handler(dpg.mvKey_Up, callback=self._press_up)
        #     dpg.add_key_press_handler(dpg.mvKey_Left, callback=self._press_left)
        #     dpg.add_key_press_handler(dpg.mvKey_Right, callback=self._press_right)

        with dpg.window(label="popup", tag="option_popup", popup=True, show=False, no_title_bar=True):
            dpg.add_button(label="Follow pointer", callback=self._popup_follow_pointer)

        with dpg.group(parent=self.parent, horizontal=True):
            dpg.add_text("Saved memory snapshots:")
            dpg.add_combo(
                items=list(self._memory_cache.keys()),
                callback=self._select_snapshot,
                tag="_snapshot_dropdown",
            )

        with dpg.group(parent=self.parent, show=False, tag="_snapshot_details_group"):
            dpg.add_text("Snapshot details:", show=False, tag="_snapshot_details_text")
            dpg.add_text("    Timestamp: N/A", tag="_snapshot_details_timestamp")
            dpg.add_button(label="Delete", callback=self._delete_snapshot)

        with dpg.group(parent=self.parent, horizontal=True):
            dpg.add_text("Address: 0x")
            dpg.add_input_text(
                source="address_input",
                callback=self._change_address,
                hexadecimal=True,
                on_enter=True,
            )

        with dpg.group(parent=self.parent, horizontal=True):
            dpg.add_text("Selection size:")
            dpg.add_input_int(
                source="selection_size",
                callback=self._change_selection_size,
                min_value=1,
                min_clamped=True,
            )

        _main_table = dpg.add_table(
            parent=self.parent,
            tag="main_table",
            header_row=True,
            borders_innerV=True,
            policy=dpg.mvTable_SizingFixedFit,
            scrollY=True,
            freeze_rows=1,
            context_menu_in_body=True,
            clipper=True,
        )

        dpg.add_table_column(
            label="Address",
            tag="col_address",
            no_resize=True,
            width_fixed=True,
            parent="main_table",
        )
        dpg.add_table_column(
            label="Hex",
            tag="col_bytes",
            parent="main_table",
        )
        dpg.add_table_column(
            label="Decoded",
            tag="col_string",
            parent="main_table",
            width_fixed=True,
        )

        self._populate_table()

        dpg.bind_item_theme(_main_table, table_theme)

        with dpg.group(parent=self.parent, horizontal=True, tag="size_2_values", show=False):
            dpg.add_text("int16:")
            dpg.add_text(source="_select_bytes_int16")
            dpg.add_text("uint16:")
            dpg.add_text(source="_select_bytes_uint16")
        with dpg.group(parent=self.parent, horizontal=True, tag="size_4_values", show=True):
            dpg.add_text("int32:")
            dpg.add_text(source="_select_bytes_int32")
            dpg.add_text("uint32:")
            dpg.add_text(source="_select_bytes_uint32")
        with dpg.group(parent=self.parent, horizontal=True, tag="size_8_values", show=False):
            dpg.add_text("int64:")
            dpg.add_text(source="_select_bytes_int64")
            dpg.add_text("uint64:")
            dpg.add_text(source="_select_bytes_uint64")
        dpg.add_text("Invalid size selected", parent=self.parent, tag="size_inv_values", show=False)

    def _setup_rightclick_handlers(self):
        with dpg.item_handler_registry(tag="byte_handler"):
            dpg.add_item_clicked_handler(dpg.mvMouseButton_Right, callback=self._clicked_on)



# This file includes various functions which use any of the various windows dlls to determine things.

import ctypes
from ctypes import wintypes
from typing import Protocol

import pymem
import pymem.ressources.structure

GetModuleFileNameExA = ctypes.windll.psapi.GetModuleFileNameExA
GetModuleFileNameExA.restype = wintypes.DWORD
GetModuleFileNameExA.argtypes = [
    wintypes.HANDLE,
    wintypes.HMODULE,
    wintypes.LPSTR,
    wintypes.DWORD,
]

GetWindowLongA = ctypes.windll.user32.GetWindowLongA
GetWindowLongA.restype = wintypes.LONG
GetWindowLongA.argtypes = [
    wintypes.HWND,
    ctypes.c_int,
]

SetWindowLongA = ctypes.windll.user32.SetWindowLongA
SetWindowLongA.restype = wintypes.LONG
SetWindowLongA.argtypes = [
    wintypes.HWND,
    ctypes.c_int,
    wintypes.LONG,
]

SetLayeredWindowAttributes = ctypes.windll.user32.SetLayeredWindowAttributes
SetLayeredWindowAttributes.restype = wintypes.BOOL
SetLayeredWindowAttributes.argtypes = [
    wintypes.HWND,
    wintypes.COLORREF,
    wintypes.BYTE,
    wintypes.DWORD,
]


SetLastError = ctypes.windll.kernel32.SetLastError
SetLastError.argtypes = [wintypes.DWORD]


GetLastError = ctypes.windll.kernel32.GetLastError
GetLastError.restype = wintypes.DWORD


VirtualQuery = ctypes.windll.kernel32.VirtualQuery
VirtualQuery.argtypes = [
    wintypes.LPCVOID,
    ctypes.c_void_p,
    ctypes.c_size_t,
]
VirtualQuery.restype = ctypes.c_size_t


MAX_EXE_NAME_SIZE = 1024
WS_EX_LAYERED = 0x00080000  # layered window
GWL_EXSTYLE = -20  # "extended window style"

LWA_COLORKEY = 0x00000001
LWA_ALPHA = 0x00000002


def get_exe_path_from_pid(proc: pymem.Pymem) -> str:
    """Get the name of the exe which was run to create the pymem process."""
    name_buffer = ctypes.create_string_buffer(b"", MAX_EXE_NAME_SIZE)
    GetModuleFileNameExA(proc.process_handle, None, name_buffer, MAX_EXE_NAME_SIZE)
    return name_buffer.value.decode()


def set_window_transparency(hwnd: int, alpha: float):
    SetWindowLongA(hwnd, GWL_EXSTYLE, GetWindowLongA(hwnd, GWL_EXSTYLE) | WS_EX_LAYERED)
    rgb = wintypes.RGB(10, 10, 10)
    SetLayeredWindowAttributes(hwnd, rgb, int(255 * alpha), LWA_ALPHA)


class MemoryInfo(Protocol):
    BaseAddress: int
    AllocationBase: int
    AllocationProtect: int
    RegionSize: int
    State: int
    Protect: int
    Type: int


def QueryAddress(address: int) -> MemoryInfo:
    mbi = pymem.ressources.structure.MEMORY_BASIC_INFORMATION()
    SetLastError(0)
    VirtualQuery(address, ctypes.byref(mbi), ctypes.sizeof(mbi))
    if errcode := GetLastError():
        raise ValueError(f"There was an error accessing address 0x{address:X}: {errcode}")
    return mbi



import asyncio
import concurrent.futures
import os
import os.path as op
import subprocess
import time
import webbrowser
from functools import partial
from signal import SIGTERM
from typing import Any, Optional

import psutil
import pymem
import pymem.exception
import pymem.process

from pymhf.core._internal import LoadTypeEnum
from pymhf.core.caching import hash_bytes
from pymhf.core.logging import open_log_console
from pymhf.core.process import start_process
from pymhf.core.protocols import ESCAPE_SEQUENCE, TerminalProtocol
from pymhf.utils.config import canonicalize_setting
from pymhf.utils.parse_toml import read_pymhf_settings
from pymhf.utils.winapi import get_exe_path_from_pid

CWD = op.dirname(__file__)
PYMHF_DIR = op.dirname(CWD)
APPDATA_DIR = os.environ.get("APPDATA", op.expanduser("~"))


class WrappedProcess:
    def __init__(self, proc: Optional[psutil.Process] = None, thread_handle: Optional[int] = None):
        self.proc = proc
        self.thread_handle = thread_handle
        self._is_self_started = False
        if self.thread_handle is not None:
            self._is_self_started = True

    def suspend(self):
        if self.proc is not None:
            self.proc.suspend()

    def resume(self):
        if self.proc is not None:
            self.proc.resume()
        else:
            pymem.ressources.kernel32.ResumeThread(self.thread_handle)


class pymhfExitException(Exception):
    pass


def get_process_when_ready(
    cmd: str,
    target: str,
    required_assemblies: Optional[list[str]] = None,
    is_steam: bool = True,
    start_paused: bool = False,
):
    target_process: Optional[WrappedProcess] = None
    parent_process = None
    # If we are running something which is under steam, make sure steam is
    # running first.
    if is_steam:
        for p in psutil.process_iter(["name", "pid"]):
            if p.name().lower() == "steam.exe":
                parent_process = p
                break
        if parent_process is None:
            raise ProcessLookupError("Steam not running! For now, start it yourself and try again...")

    run = True
    if parent_process is not None:
        if is_steam:
            # For steam games, if we have the game ID then the cmd will be
            # steam://rungameid/{game id} which we need to invoke this way.
            print("Running from steam")
            webbrowser.open(cmd)
        else:
            subprocess.run(cmd)
        while run:
            try:
                if target_process is None:
                    for child in parent_process.children(recursive=True):
                        if child.name() == target:
                            target_process = WrappedProcess(proc=child)
                else:
                    binary = pymem.Pymem(target)
                    modules = list(pymem.process.enum_process_module(binary.process_handle))
                    if len(modules) > 0:
                        if set(required_assemblies) <= set(x.name for x in modules):
                            run = False
                            break
            except KeyboardInterrupt:
                raise
    else:
        creationflags = 0x4 if start_paused else 0
        process_handle, thread_handle, pid, tid = start_process(  # noqa
            cmd, creationflags=creationflags
        )
        target_process = WrappedProcess(thread_handle=thread_handle)

    if target_process is not None:
        binary = pymem.Pymem(target)
        # TODO: We can inject python in really early since any loaded dll's will persist through steam's
        # forking process.
        try:
            binary.inject_python_interpreter()
        except pymem.exception.MemoryWriteError:
            print("Failed to inject python for some reason... Trying again in 2 seconds")
            time.sleep(2)
            binary.inject_python_interpreter()
        print(f"Python injected into pid {binary.process_id}")
        if start_paused:
            target_process.suspend()

        return binary, target_process
    else:
        return None, None


def load_mod_file(filepath):
    """Load an individual file as a mod."""
    pymhf_settings = read_pymhf_settings(filepath, True)
    _run_module(filepath, pymhf_settings, None, None)


def load_module(plugin_name: str, module_path: str):
    """Load the module."""
    config_dir = op.join(APPDATA_DIR, "pymhf", plugin_name)
    local_cfg_file = op.join(config_dir, "pymhf.local.toml")
    if op.exists(local_cfg_file):
        local_cfg = read_pymhf_settings(local_cfg_file).get("local_config", {})
    else:
        local_cfg = {}
    module_cfg_file = op.join(module_path, "pymhf.toml")
    module_cfg = read_pymhf_settings(module_cfg_file)
    module_cfg.update(local_cfg)

    _run_module(module_path, module_cfg, plugin_name, config_dir)


def _required_config_val(config: dict[str, str], key: str) -> Any:
    if (val := config.get(key)) is not None:
        return val
    raise ValueError(f"[tool.pymhf] missing config value: {key}")


def _run_module(module_path: str, config: dict[str, str], plugin_name: Optional[str], config_dir: str):
    """Run the module provided.

    Parameters
    ----------
    module_path
        The path to the module or single-file mod to be run.
    config
        A mapping of the associated pymhf config.
    """
    load_type = LoadTypeEnum.LIBRARY
    if op.isfile(module_path):
        load_type = LoadTypeEnum.SINGLE_FILE
    if op.exists(op.join(module_path, "pymhf.toml")):
        load_type = LoadTypeEnum.MOD_FOLDER

    binary_path = None
    binary_exe = _required_config_val(config, "exe")
    required_assemblies = []
    required_assemblies = config.get("required_assemblies", [])
    start_exe = config.get("start_exe", True)
    try:
        steam_gameid = int(config.get("steam_gameid", 0))
    except (ValueError, TypeError):
        steam_gameid = 0
    start_paused = config.get("start_paused", False)

    if config_dir is None:
        cache_dir = op.join(module_path, ".cache")
    else:
        cache_dir = op.join(config_dir, ".cache")

    # Check if the module_path is a file or a folder.
    _module_path = module_path
    if op.isfile(module_path):
        _module_path = op.dirname(module_path)

    is_steam = False
    if steam_gameid:
        cmd = f"steam://rungameid/{steam_gameid}"
        is_steam = True
    else:
        if op.isabs(binary_exe):
            cmd = binary_path = binary_exe
            # We only need the binary_exe to be the name from here on.
            binary_exe = op.basename(binary_exe)
        else:
            # TODO: Allow support for running local binaries or binaries which are on the path.
            raise ValueError("[tool.pymhf].exe must be a full path if no steam_gameid is provided.")

    if start_exe:
        pm_binary, proc = get_process_when_ready(cmd, binary_exe, required_assemblies, is_steam, start_paused)
    else:
        # If we aren't starting the exe then by the time we have run we expect the process to already exist,
        # so just find it with pymem.
        pm_binary = pymem.Pymem(binary_exe)
        pm_binary.inject_python_interpreter()
        print("Python injected")
        if binary_path is None:
            binary_path = get_exe_path_from_pid(pm_binary)
        proc = None

    if not pm_binary and not proc:
        # TODO: Raise better error messages/reason why it couldn't load.
        print("FATAL ERROR: Cannot start process!")
        return

    # When we start from steam, the binary path will be None, so retreive it from from the psutils Process
    # object.
    if binary_path is None and proc is not None:
        binary_path = proc.proc.exe()

    binary_dir = None
    if binary_path is not None:
        binary_dir = op.dirname(binary_path)

    print(f"Found PID: {pm_binary.process_id}")

    logging_config = config.get("logging", {})
    log_window_name_override = logging_config.get("window_name_override", "pymhf console")
    log_dir = logging_config.get("log_dir", "{CURR_DIR}")
    log_dir = canonicalize_setting(log_dir, plugin_name, _module_path, binary_dir, "LOGS")

    mod_save_dir = config.get("mod_save_dir", "{CURR_DIR}")
    mod_save_dir = canonicalize_setting(mod_save_dir, plugin_name, _module_path, binary_dir, "MOD_SAVES")

    executor = None
    futures = []
    loop = asyncio.get_event_loop()

    def kill_injected_code(loop: asyncio.AbstractEventLoop):
        # End one last "escape sequence" message:
        client_completed = asyncio.Future()
        client_factory = partial(TerminalProtocol, message=ESCAPE_SEQUENCE.decode(), future=client_completed)
        factory_coroutine = loop.create_connection(
            client_factory,
            "127.0.0.1",
            6770,
        )
        loop.run_until_complete(factory_coroutine)
        loop.run_until_complete(client_completed)

    try:
        log_pid = open_log_console(op.join(CWD, "log_terminal.py"), log_dir, log_window_name_override)
        # Have a small nap just to give it some time.
        time.sleep(0.5)
        print(f"Opened the console log with PID: {log_pid}")
        if binary_path is not None:
            with open(binary_path, "rb") as f:
                binary_hash = hash_bytes(f)
            print(f"Exe hash is: {binary_hash}")
        else:
            binary_hash = 0

        def close_callback(x):
            print("pyMHF exiting...")
            for _pid in {pm_binary.process_id, log_pid}:
                try:
                    os.kill(_pid, SIGTERM)
                except Exception:
                    # If we can't kill it, it's probably already dead. Just continue.
                    pass
            # Finally, send a SIGTERM to ourselves...
            os.kill(os.getpid(), SIGTERM)

        # Wait some time for the data to be written to memory.
        time.sleep(3)

        print(f"proc id from pymem: {pm_binary.process_id}")
        offset_map = {}
        # This is a mapping of the required assemblies to their filenames so we may do an import look up on
        # the inside.
        included_assemblies = {}
        modules = []
        if required_assemblies:
            modules = list(pm_binary.list_modules())
            found_modules = list(filter(lambda x: x.name in required_assemblies, modules))
            if not found_modules:
                print(f"Cannot find specified assembly from config ({required_assemblies})")
                return
            for module in found_modules:
                offset_map[module.name] = (module.lpBaseOfDll, module.SizeOfImage)
                included_assemblies[module.name] = module.filename
        else:
            pb = pm_binary.process_base
            offset_map[binary_exe] = (pb.lpBaseOfDll, pb.SizeOfImage)
        # if not modules:
        #     modules = pm_binary.list_modules()
        # for module in modules:
        #     included_assemblies[module.name] = module.filename
        # print(f"The handle: {pm_binary.process_handle}, bases: {offset_map}")

        if len(offset_map) == 1:
            # Only one thing. For now, we just set these as the `binary_base` and `binary_size`.
            # TODO: When we want to support offsets and hooks in multiple assemblies we need to pass the whole
            # dictionary in potentially, or do a lookup from inside the process.
            assem_name = list(offset_map.keys())[0]
            binary_base = offset_map[assem_name][0]
            binary_size = offset_map[assem_name][1]

        try:
            cwd = CWD.replace("\\", "\\\\")
            import sys

            _path = sys.path
            _path.insert(0, PYMHF_DIR)

            saved_path = [x.replace("\\", "\\\\") for x in _path]
            # TODO: This can fail sometimes.... Figure out why??
            pm_binary.inject_python_shellcode(
                f"""
import sys
sys.path = {saved_path}
                """
            )

            with open(op.join(CWD, "_preinject.py"), "r") as f:
                _preinject_shellcode = f.read()
            pm_binary.inject_python_shellcode(_preinject_shellcode)
            # Inject the common NMS variables which are required for general use.
            module_path = op.realpath(module_path)
            module_path = module_path.replace("\\", "\\\\")
            pm_binary.inject_python_shellcode(
                f"""
pymhf.core._internal.MODULE_PATH = {module_path!r}
pymhf.core._internal.BASE_ADDRESS = {binary_base!r}
pymhf.core._internal.SIZE_OF_IMAGE = {binary_size!r}
pymhf.core._internal.CWD = {cwd!r}
pymhf.core._internal.PID = {pm_binary.process_id!r}
pymhf.core._internal.HANDLE = {pm_binary.process_handle!r}
pymhf.core._internal.BINARY_HASH = {binary_hash!r}
pymhf.core._internal.CONFIG = {config!r}
pymhf.core._internal.EXE_NAME = {binary_exe!r}
pymhf.core._internal.BINARY_PATH = {binary_path!r}
pymhf.core._internal.LOAD_TYPE = {load_type.value!r}
pymhf.core._internal.MOD_SAVE_DIR = {mod_save_dir!r}
pymhf.core._internal.INCLUDED_ASSEMBLIES = {included_assemblies!r}
pymhf.core._internal.CACHE_DIR = {cache_dir!r}
                """
            )
        except Exception as e:
            print(e)
        # Inject the script
        with open(op.join(CWD, "injected.py"), "r") as f:
            shellcode = f.read()
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=2)
        print(f"Injecting hooking code into proc id {pm_binary.process_id}")
        fut = executor.submit(pm_binary.inject_python_shellcode, shellcode)
        fut.add_done_callback(lambda x: close_callback(x))
        futures.append(fut)

        # Wait for a user input to start the process.
        # TODO: Send a signal back up from the process to trigger this automatically.
        if start_paused and start_exe is not False:
            try:
                #     # print("Checking to see if we are ready to run...")
                #     # client_completed = asyncio.Future()
                #     # client_factory = partial(
                #     #     TerminalProtocol,
                #     #     message=READY_ASK_SEQUENCE,
                #     #     future=client_completed
                #     # )
                #     # print("A")
                #     # factory_coroutine = loop.create_connection(
                #     #     client_factory,
                #     #     '127.0.0.1',
                #     #     6770,
                #     # )
                #     # print("B")
                #     # loop.run_until_complete(factory_coroutine)
                #     # print("C")
                #     # loop.run_until_complete(client_completed)

                input("Press something to start binary")
            except KeyboardInterrupt:
                # Kill the injected code so that we don't wait forever for the future to end.
                kill_injected_code(loop)
                raise
            proc.resume()

        print("pyMHF interactive python command prompt")
        print("Type any valid python commands to execute them within the games' process")
        # TODO: change this to use a threading.Event object, and in a separate thread, poll every second or so
        # the running game process to see if it's still running (using psutil?)
        # Might need to modify the WrappedProcess object to always create the pstuil.Process object to make it
        # work easier...

        # TODO: This should become False when we exit form the program...
        while True:
            try:
                input_ = input(">>> ")
                client_completed = asyncio.Future()
                client_factory = partial(TerminalProtocol, message=input_, future=client_completed)
                factory_coroutine = loop.create_connection(
                    client_factory,
                    "127.0.0.1",
                    6770,
                )
                loop.run_until_complete(factory_coroutine)
                loop.run_until_complete(client_completed)
            except KeyboardInterrupt:
                kill_injected_code(loop)
                raise
    except KeyboardInterrupt:
        # If it's a keyboard interrupt, just pass as it will have bubbled up from
        # below.
        pass
    except Exception as e:
        # Any other exception we want to actually know about.
        print(e)
        raise
    finally:
        loop.close()
        try:
            for future in concurrent.futures.as_completed(futures, timeout=5):
                print(future)
        except TimeoutError:
            # Don't really care.
            print("Got a time out error...")
            pass
        if executor is not None:
            executor.shutdown(wait=False)
        try:
            with open(op.join(CWD, "end.py"), "r") as f:
                close_shellcode = f.read()
            pm_binary.inject_python_shellcode(close_shellcode)
            print("Just injected the close command?")
            # Kill the process.
        except Exception:
            pass
        finally:
            print("Forcibly shutting down process")
            time.sleep(1)
            for _pid in {pm_binary.process_id, log_pid}:
                try:
                    os.kill(_pid, SIGTERM)
                    print(f"Just killed process {_pid}")
                except Exception:
                    # If we can't kill it, it's probably already dead. Just continue.
                    print(f"Failed to kill process {_pid}. It was likely already dead...")
                    pass



# Main functionality for loading mods.

# Mods will consist of a single file which will generally contain a number of
# hooks.

import ctypes
import importlib
import importlib.util
import inspect
import json
import logging
import os
import os.path as op
import sys
import traceback
from abc import ABC
from dataclasses import fields
from functools import partial
from types import ModuleType
from typing import TYPE_CHECKING, Any, Optional, Type, TypeVar, Union

import keyboard
from packaging.version import InvalidVersion
from packaging.version import parse as parse_version

import pymhf.core._internal as _internal
from pymhf.core._types import HookProtocol, KeyPressProtocol
from pymhf.core.errors import NoSaveError
from pymhf.core.hooking import HookManager
from pymhf.core.importing import import_file
from pymhf.core.memutils import get_addressof, map_struct
from pymhf.core.module_data import module_data
from pymhf.core.utils import does_pid_have_focus, saferun
from pymhf.gui.protocols import ButtonProtocol, ComboBoxProtocol, VariableProtocol

if TYPE_CHECKING:
    from pymhf.gui.gui import GUI


mod_logger = logging.getLogger("ModManager")


def _is_mod_predicate(obj, ref_module) -> bool:
    if inspect.getmodule(obj) == ref_module and inspect.isclass(obj):
        return issubclass(obj, Mod) and getattr(obj, "_should_enable", True)
    return False


def _is_mod_state_predicate(obj) -> bool:
    return isinstance(obj, ModState)


def _funchook_predicate(value: Any) -> bool:
    return getattr(value, "_is_funchook", False)


def _callback_predicate(value: Any) -> bool:
    return hasattr(value, "_custom_trigger")


def _has_hotkey_predicate(value: Any) -> bool:
    """Determine if the object has the _is_main_loop_func property.
    This will only be methods on Mod classes which are decorated with either
    @main_loop.before or @main_loop.after
    """
    return getattr(value, "_hotkey", False)


def _gui_button_predicate(value) -> bool:
    return getattr(value, "_is_button", False) and hasattr(value, "_button_text")


def _gui_combobox_predicate(value) -> bool:
    return getattr(value, "_is_combobox", False) and hasattr(value, "_combobox_text")


def _gui_variable_predicate(value) -> bool:
    # Variables are properties which have the .fset function defined with
    # _is_variable and _label_text attributes, and a .fset function too.
    if isinstance(value, property):
        return (
            getattr(value.fget, "_is_variable", False)
            and hasattr(value.fget, "_label_text")
            and hasattr(value.fget, "_variable_type")
        )


class StructEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, "__json__"):
            return {
                "struct": obj.__class__.__qualname__,
                "module": obj.__class__.__module__,
                "fields": obj.__json__(),
            }
        return json.JSONEncoder.default(self, obj)


class StructDecoder(json.JSONDecoder):
    def __init__(self):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook)

    def object_hook(seld, obj: dict):
        if (module := obj.get("module")) is not None:
            if module == "__main__":
                return globals()[obj["struct"]](**obj["fields"])
            else:
                try:
                    module_ = importlib.import_module(module)
                    return getattr(module_, obj["struct"])(**obj["fields"])
                except ImportError:
                    mod_logger.error(f"Cannot import {module}")
                    return
                except AttributeError:
                    mod_logger.error(f"Cannot find {obj['struct']} in {module}")
                    return
        return obj


class ModState(ABC):
    """A class which is used as a base class to indicate that the class is to be used as a mod state.

    Mod State classes will persist across mod reloads so any variables set in it
    will have the same value after the mod has been reloaded.
    """

    _save_fields_: tuple[str]

    def save(self, name: str):
        """Save the current mod state to a file.

        Parameters
        ----------
        name:
            The name of the file this ``ModState`` will be saved to.
            Note that this will be a json file saved within the ``MOD_SAVE_DIR`` directory.
        """
        _data = {}
        if hasattr(self, "_save_fields_") and self._save_fields_:
            for field in self._save_fields_:
                _data[field] = getattr(self, field)
        else:
            try:
                for f in fields(self):
                    _data[f.name] = getattr(self, f.name)
            except TypeError:
                mod_logger.error(
                    "To save a mod state it must either be a dataclass or "
                    "have the _save_fields_ attribute. State was not saved"
                )
                return
        if not _internal.MOD_SAVE_DIR:
            _internal.MOD_SAVE_DIR = op.join(_internal.MODULE_PATH, "MOD_SAVES")
            mod_logger.warning(
                f"No mod_save_dir config value set. Please set one. Falling back to {_internal.MOD_SAVE_DIR}"
            )
        if not op.exists(_internal.MOD_SAVE_DIR):
            os.makedirs(_internal.MOD_SAVE_DIR)
        with open(op.join(_internal.MOD_SAVE_DIR, name), "w") as fobj:
            json.dump(_data, fobj, cls=StructEncoder, indent=1)

    def load(self, name: str):
        """Load the mod state from a file.

        Parameters
        ----------
        name:
            The name of the file this ``ModState`` will be loaded from.
            Note that this will be a json file loaded from the ``MOD_SAVE_DIR`` directory.
        """
        try:
            with open(op.join(_internal.MOD_SAVE_DIR, name), "r") as f:
                data = json.load(f, cls=StructDecoder)
        except FileNotFoundError as e:
            raise NoSaveError from e
        for key, value in data.items():
            setattr(self, key, value)


class Mod(ABC):
    __author__: Union[str, list[str]] = "Name(s) of the mod author(s)"
    __description__: str = "Short description of the mod"
    __version__: str = "Mod version"
    __dependencies__: list[str] = []
    # Minimum required pyMHF version for this mod.
    __pymhf_required_version__: Optional[str] = None

    _custom_callbacks: dict[str, set[HookProtocol]]
    pymhf_gui: "GUI"

    def __init__(self):
        # Find all the hooks defined for the mod.
        self.hooks: set[HookProtocol] = self.get_members(_funchook_predicate)
        self._custom_callbacks = self.get_members(_callback_predicate)
        self._hotkey_funcs: set[KeyPressProtocol] = self.get_members(_has_hotkey_predicate)
        self._gui_buttons: dict[str, ButtonProtocol] = {
            x[1].__qualname__: x[1] for x in inspect.getmembers(self, _gui_button_predicate)
        }
        self._gui_comboboxes: dict[str, ComboBoxProtocol] = {
            x[1].__qualname__: x[1] for x in inspect.getmembers(self, _gui_combobox_predicate)
        }
        self.pymhf_gui = None
        # For variables, unless there is a better way, store just the name so we
        # can our own special binding of the name to the GUI.
        self._gui_variables: dict[str, VariableProtocol] = {}
        for x in inspect.getmembers(self.__class__, _gui_variable_predicate):
            if x[1].fset is not None:
                x[1].fget._has_setter = True
            self._gui_variables[x[0]] = x[1].fget

    @property
    def _mod_name(self):
        return self.__class__.__name__

    def get_members(self, predicate):
        return {x[1] for x in inspect.getmembers(self, predicate)}


T = TypeVar("T", bound=Mod)


class ModManager:
    def __init__(self):
        # Internal mapping of mods.
        # TODO: Probably change datatype
        self._preloaded_mods: dict[str, type[Mod]] = {}
        # Actual mapping of mods.
        self.mods: dict[str, Mod] = {}
        self._mod_hooks: dict[str, list] = {}
        self.mod_states: dict[str, list[tuple[str, ModState]]] = {}
        self._mod_paths: dict[str, ModuleType] = {}
        self.hook_manager: HookManager = None
        # Keep a mapping of the hotkey callbacks
        self.hotkey_callbacks: dict[tuple[str, str], Any] = {}

    def __getitem__(self, key: Type[T]) -> T:
        if not issubclass(key, Mod):
            raise TypeError("The lookup object must be the class type")
        return self.mods.get(key.__name__)

    def _load_module(self, module: ModuleType) -> bool:
        """Load a mod from the provided module.

        This will be called when initially loading the mods, and also when we
        wish to reload a mod.
        """
        d: dict[str, type[Mod]] = dict(
            inspect.getmembers(module, partial(_is_mod_predicate, ref_module=module))
        )
        if not len(d) >= 1:
            mod_logger.error(
                f"The file {module.__file__} has more than one mod defined in it. "
                "Only define one mod per file."
            )
        if len(d) == 0:
            # No mod in the file. Just return
            return False
        mod_name = list(d.keys())[0]
        mod = d[mod_name]
        if mod.__pymhf_required_version__ is not None:
            from pymhf import __version__ as _pymhf_version

            try:
                pymhf_version = parse_version(_pymhf_version)
            except InvalidVersion:
                pymhf_version = None
            try:
                mod_version = parse_version(mod.__pymhf_required_version__)
            except InvalidVersion:
                mod_logger.warning(
                    f"__pymhf_required_version__ defined on mod {mod.__name__} is not a valid version string"
                )
                mod_version = None
            if mod_version is None or mod_version <= pymhf_version:
                self._preloaded_mods[mod_name] = mod
            else:
                mod_logger.error(
                    f"Mod {mod.__name__} requires a newer verison of "
                    f"pyMHF ({mod_version} ≥ {pymhf_version})! "
                    "Please update"
                )
        else:
            self._preloaded_mods[mod_name] = mod
        # Only get mod states if the mod name doesn't already have a cached
        # state, otherwise it will override it.
        if mod_name not in self.mod_states:
            mod_states = list(inspect.getmembers(mod, _is_mod_state_predicate))
            self.mod_states[mod_name] = mod_states
        if not mod_name.startswith("_INTERNAL_"):
            self._mod_paths[mod_name] = module
        return True

    def load_mod(self, fpath) -> Optional[ModuleType]:
        """Load a mod from the given filepath.

        This returns the loaded module if it contains a valid mod and can be loaded correctly.
        """
        module = import_file(fpath)
        if module is None:
            return None
        if self._load_module(module):
            return module

    # TODO: Can probably move the duplicated functionality between this and the next method into a single
    # function.
    def load_single_mod(self, fpath: str, bind: bool = True):
        """Load a single mod file.

        Params
        ------
        folder
            The path of the folder to be loaded. All mod files within this directory will be loaded and
            installed.
        bind
            Whether or not to actual bind and initialize the hooks within the mod.
            This should almost always be True except when loading the internal mods initially since it's not
            necessary.
            If this function is called with False, then it MUST be called again with True before the hooks
            are enabled.
        """
        self.load_mod(fpath)
        # Once all the mods in the folder have been loaded, then parse the mod
        # for function hooks and register then with the hook loader.
        loaded_mods = len(self._preloaded_mods)
        for _mod in self._preloaded_mods.values():
            self.instantiate_mod(_mod)

        self._preloaded_mods.clear()

        bound_hooks = 0
        if bind:
            bound_hooks = self.hook_manager.initialize_hooks()

        return loaded_mods, bound_hooks

    def load_mod_folder(self, folder: str, bind: bool = True) -> tuple[int, int]:
        """Load the mod folder.

        Params
        ------
        folder
            The path of the folder to be loaded. All mod files within this directory will be loaded and
            installed.
        bind
            Whether or not to actual bind and initialize the hooks within the mod.
            This should almost always be True except when loading the internal mods initially since it's not
            necessary.
            If this function is called with False, then it MUST be called again with True before the hooks
            are enabled.
        """
        for file in os.listdir(folder):
            if file.endswith(".py"):
                self.load_mod(op.join(folder, file))
        # Once all the mods in the folder have been loaded, then parse the mod
        # for function hooks and register then with the hook loader.
        loaded_mods = len(self._preloaded_mods)
        for _mod in self._preloaded_mods.values():
            self.instantiate_mod(_mod)

        self._preloaded_mods.clear()

        bound_hooks = 0
        if bind:
            bound_hooks = self.hook_manager.initialize_hooks()

        return loaded_mods, bound_hooks

    def instantiate_mod(self, mod: type[Mod], quiet: bool = False) -> Mod:
        """Register all the functions within the mod as hooks."""
        _mod = mod()
        # First register each of the methods which are detours.
        for hook in _mod.hooks:
            self.hook_manager.register_hook(hook)
        # Add any custom callbacks which may be defined by the calling library.
        self.hook_manager._add_custom_callbacks(_mod._custom_callbacks)
        # Finally, set up any keyboard bindings.
        for hotkey_func in _mod._hotkey_funcs:
            # Don't need to tell the hook manager, register the keyboard
            # hotkey here...
            # NOTE: The below is a "hack"/"solution" to an issue that the
            # keyboard library has.
            # cf. https://github.com/boppreh/keyboard/issues/584
            # NOTE: This solution breaks for keybindings where multiple keys
            # Are required. Will need a better solution for this case.

            cb = keyboard.hook(
                lambda e, func=hotkey_func, name=hotkey_func._hotkey, event_type=hotkey_func._hotkey_press: (
                    e.name == name
                    and e.event_type == event_type
                    and does_pid_have_focus(_internal.PID)
                    and saferun(func)
                )
            )
            self.hotkey_callbacks[(hotkey_func._hotkey, hotkey_func._hotkey_press)] = cb
        self.mods[_mod._mod_name] = _mod
        return _mod

    def _gui_reload(self, _sender, _keyword, user_data: tuple[str, "GUI"]):
        # Callback to register with the GUI to enable reloading of mods from there.
        self.reload(*user_data)
        self._assign_mod_instances(user_data[0])

    def _assign_mod_instances(self, specific_mod: Optional[str] = None):
        """Assign the types of the mod classes in each mod to all of the mods which are loaded."""
        # Loop over the loaded mods. If it has any dependencies, get the module it belongs to and assign those
        # dependencies to it.
        if specific_mod is not None:
            iter_ = [(specific_mod, self.mods[specific_mod])]
        else:
            iter_ = self.mods.items()
        for _mod_name, mod in iter_:
            if dependencies := getattr(mod, "__dependencies__", []):
                module = self._mod_paths[_mod_name]
                for dependency in dependencies:
                    setattr(module, dependency, self.mods[dependency].__class__)

    def reload(self, name: str, gui: "GUI"):
        """Reload a mod with the given name."""
        try:
            if (mod := self.mods.get(name)) is not None:
                # First, remove everything.
                for hook in mod.hooks:
                    hook_name = hook._hook_func_name
                    fh = self.hook_manager.hooks.get(hook_name)
                    if fh is not None:
                        mod_logger.info(f"Removing hook {hook}: {hook._hook_func_name}")
                        fh.remove_detour(hook)
                        # If the hook has been closed it means that there are now no longer any methods
                        # assigned as detours to it. Remove the hook from the registry.
                        if fh.state == "closed":
                            del self.hook_manager.hooks[hook_name]

                self.hook_manager._remove_custom_callbacks(mod._custom_callbacks)
                for hotkey_func in mod._hotkey_funcs:
                    cb = self.hotkey_callbacks.pop(
                        (hotkey_func._hotkey, hotkey_func._hotkey_press),
                        None,
                    )
                    if cb is not None:
                        keyboard.unhook(cb)

                # Then, reload the module.
                module = self._mod_paths[name]
                del sys.modules[module.__name__]

                # Then, add everything back.
                new_module = self.load_mod(module.__file__)
                for _mod in self._preloaded_mods.values():
                    mod = self.instantiate_mod(_mod)

                    # Get the mod states for the mod if there are any and reapply them to the new mod
                    # instance.
                    if mod_state := self.mod_states.get(name):
                        for ms in mod_state:
                            field, state = ms
                            member_req_reinst = {}
                            for x in inspect.getmembers(state):
                                member, member_type = x
                                if not member.startswith("__"):
                                    if (
                                        _module := getattr(member_type.__class__, "__module__", None)
                                    ) is not None and isinstance(member_type, ctypes.Structure):
                                        if _module == module.__spec__.name:
                                            # In this case, the instance of the attribute in the ModState was
                                            # defined in the module that is being reloaded. We need to
                                            # re-instantiate it so that we can get any potential changes to
                                            # it.
                                            member_req_reinst[member] = member_type
                                            logging.info(f"{member}: {_module}")
                            logging.info(
                                f"Reinstantiating the following members: {list(member_req_reinst.keys())}"
                            )
                            for name, type_ in member_req_reinst.items():
                                data_offset = get_addressof(type_)
                                new_obj_type_name = type_.__class__.__name__
                                mod_logger.info(f"{name} is of type {new_obj_type_name}")
                                new_obj_type = getattr(new_module, new_obj_type_name)
                                new_obj = map_struct(data_offset, new_obj_type)
                                setattr(state, name, new_obj)
                                del member_type
                            setattr(mod, field, state)

                    # Check also to see if the file had any module-level __pymhf attributes which we might
                    # to update the `module_data` with.
                    _new_binary = getattr(new_module, "__pymhf_func_binary__", None)
                    _new_offsets = getattr(new_module, "__pymhf_func_offsets__", {})
                    _new_patterns = getattr(new_module, "__pymhf_func_patterns__", {})
                    _new_func_call_sigs = getattr(new_module, "__pymhf_func_call_sigs__", {})

                    if _new_binary:
                        module_data.FUNC_BINARY = _new_binary
                    module_data.FUNC_OFFSETS.update(_new_offsets)
                    module_data.FUNC_PATTERNS.update(_new_patterns)
                    module_data.FUNC_CALL_SIGS.update(_new_func_call_sigs)

                    # Return to GUI land to reload the mod.
                    gui.reload_tab(mod)

                self._preloaded_mods.clear()

                self.hook_manager.initialize_hooks()

                # TODO: Add ability to check whether the attributes of the mod state have changed. If so,
                # remove or add these attributes as required. Might want to have some kind of "copy" method to
                # actually create a new instance each time but persist the data across.

                mod_logger.info(f"Finished reloading {name}")
            else:
                mod_logger.error(f"Cannot find mod {name}")
        except Exception:
            mod_logger.error(traceback.format_exc())


mod_manager = ModManager()



import ctypes
from dataclasses import dataclass
from typing import Annotated, Optional, Type, TypeVar, cast

StructType = TypeVar("StructType", bound=Type[ctypes.Structure])


@dataclass
class Field:
    datatype: ctypes.Structure
    offset: Optional[int] = None


def partial_struct(cls: StructType):
    _fields_ = []
    curr_position = 0
    for field_name, annotation in cls.__annotations__.items():
        if len(annotation.__metadata__) == 1:
            field_data = annotation.__metadata__[0]
        else:
            raise ValueError(f"The field {field_name} has an invalid annotation: {annotation}")
        field_data = cast(Field, field_data)
        field_type = field_data.datatype
        field_offset = field_data.offset
        if field_offset and field_offset > curr_position:
            padding_bytes = field_offset - curr_position
            _fields_.append((f"_padding_{curr_position:X}", ctypes.c_ubyte * padding_bytes))
            curr_position += padding_bytes
        _fields_.append((field_name, field_type))
        curr_position += ctypes.sizeof(field_type)
    cls._fields_ = _fields_
    return cls


if __name__ == "__main__":

    @partial_struct
    class Test(ctypes.Structure):
        a: Annotated[int, Field(ctypes.c_uint32)]
        b: Annotated[int, Field(ctypes.c_uint32, 0x10)]

    print(Test._fields_)
    data = bytearray(b"\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00\x04\x00\x00\x00\x05\x00\x00\x00")
    t = Test.from_buffer(data)
    print(t.a)
    print(t.b)

    print(bytes(t))



class UnknownFunctionError(Exception):
    pass


class HookError(Exception):
    def __init__(self, status, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.status = status


class NoSaveError(Exception):
    pass



# This file contains a number of things to import or set before anything else is injected.
# For the most part this imports stuff so that the import cache is primed so that any subsequent imports do
# not overwrite the existing data.
import pymhf.core._internal  # noqa



import ctypes
import os.path as op
import sys
from logging import getLogger

import pefile

imports = {}

logger = getLogger(__name__)


SPHINX_AUTODOC_RUNNING = "sphinx.ext.autodoc" in sys.modules


def get_imports(binary_path: str) -> dict[str, ctypes._CFuncPtr]:
    directory, binary = op.split(binary_path)
    pe = pefile.PE(op.join(directory, binary), fast_load=True)
    pe.parse_data_directories(directories=[pefile.DIRECTORY_ENTRY["IMAGE_DIRECTORY_ENTRY_IMPORT"]])
    if not hasattr(pe, "DIRECTORY_ENTRY_IMPORT"):
        return {}
    for entry in pe.DIRECTORY_ENTRY_IMPORT:
        dll_name: str = entry.dll.decode()
        if dll_name.lower().endswith(".dll"):
            dll_name = dll_name.lower()[:-4]
        dll_imports = []
        for imp in entry.imports:
            if imp.name:
                dll_imports.append(imp.name.decode())
        imports[dll_name] = dll_imports

    funcptrs = {}
    for _dll, dll_imports in imports.items():
        try:
            dll = ctypes.WinDLL(_dll)
        except FileNotFoundError:
            # Re-add .dll to the end since absolute paths require the full path.
            fullpath = op.join(directory, _dll) + ".dll"
            try:
                dll = ctypes.WinDLL(fullpath)
            except FileNotFoundError:
                logger.error(f"Cannot find dll {_dll} (even at {fullpath})")
                continue
        _funcptrs = {}
        for name in dll_imports:
            func_ptr = getattr(dll, name, None)
            if func_ptr is None:
                logger.error(f"Cannot find function {_dll}.{name}")
                continue
            _funcptrs[name] = func_ptr
        funcptrs[_dll] = _funcptrs

    return funcptrs
